# Implementation based on tf.keras.engine.data_adapter.py
# https://github.com/tensorflow/tensorflow/blob/2b96f3662bd776e277f86997659e61046b56c315/tensorflow/python/keras/engine/data_adapter.py


import contextlib
import logging

from .array_adapter import ArrayDataAdapter
from .generator_adapter import GeneratorDataAdapter
from .list_adapter import ListsOfScalarsDataAdapter
from .dataset import DataLoaderAdapter

try:
    from .tf_dataset_adapter import TFDatasetAdapter
except ImportError:
    TFDatasetAdapter = None
try:
    from .torch_dataloader_adapter import TorchDataLoaderAdapter
except ImportError:
    TorchDataLoaderAdapter = None

ALL_ADAPTER_CLS = [
    ArrayDataAdapter,
    GeneratorDataAdapter,
    ListsOfScalarsDataAdapter,
    DataLoaderAdapter,
]

if TFDatasetAdapter is not None:
    ALL_ADAPTER_CLS.append(TFDatasetAdapter)
if TorchDataLoaderAdapter is not None:
    ALL_ADAPTER_CLS.append(TorchDataLoaderAdapter)


class DataHandler(object):
    """Handles iterating over epoch-level `tp.Iterator` objects."""

    def __init__(
        self,
        x,
        y=None,
        sample_weight=None,
        batch_size=None,
        steps_per_epoch=None,
        initial_epoch=0,
        epochs=1,
        shuffle=False,
        class_weight=None,
        **kwargs,
    ):

        self._initial_epoch = initial_epoch
        self._epochs = epochs
        self._insufficient_data = False

        adapter_cls = select_data_adapter(x, y)
        self._adapter = adapter_cls(
            x,
            y,
            batch_size=batch_size,
            steps=steps_per_epoch,
            epochs=epochs - initial_epoch,
            sample_weights=sample_weight,
            shuffle=shuffle,
            **kwargs,
        )

        dataset = self._adapter.get_dataset()

        self._inferred_steps = self._infer_steps(steps_per_epoch, dataset)
        self._dataset = dataset

    def enumerate_epochs(self):
        """Yields `(epoch, tp.Iterator)`."""
        data_iterator = self._dataset()
        for epoch in range(self._initial_epoch, self._epochs):
            if self._insufficient_data:  # Set by `catch_stop_iteration`.
                break
            if self._adapter.should_recreate_iterator():
                data_iterator = self._dataset()
            yield epoch, data_iterator
            self._adapter.on_epoch_end()

    @contextlib.contextmanager
    def catch_stop_iteration(self):
        """Catches errors when an iterator runs out of data."""
        try:
            yield
            # context.async_wait()

        except (StopIteration):
            if (
                self._adapter.get_size() is None
                and self._inferred_steps is None
                and self._current_step > 0
            ):
                # The input passed by the user ran out of batches.
                # Now we know the cardinality of the input(dataset or generator).
                self._inferred_steps = self._current_step
            else:
                self._insufficient_data = True
                total_epochs = self._epochs - self._initial_epoch
                logging.warning(
                    "Your input ran out of data; interrupting training. "
                    "Make sure that your dataset or generator can generate at "
                    "least `steps_per_epoch * epochs` batches (in this case, "
                    "{} batches). You may need to use the repeat() function "
                    "if using tf.data.Dataset.".format(
                        total_epochs * self._inferred_steps
                    )
                )

    def steps(self):
        """Yields steps for the current epoch."""
        self._current_step = 0
        # `self._inferred_steps` can be changed by `catch_stop_iteration`.
        while self._inferred_steps is None or self._current_step < self._inferred_steps:
            if self._insufficient_data:  # Set by `catch_stop_iteration`.
                break
            yield self._current_step
            self._current_step += 1

    @property
    def inferred_steps(self):
        """The inferred steps per epoch of the created `Dataset`.
        This will be `None` in the case where:
        (1) A generator `Dataset` was passed to the `DataHandler`, and
        (2) `steps_per_epoch` was not provided, and
        (3) The first epoch of iteration has not yet completed.
        Returns:
        The inferred steps per epoch of the created `Dataset`.
        """
        return self._inferred_steps

    def _infer_steps(self, steps, dataset):
        """Infers steps_per_epoch needed to loop through a dataset."""
        if steps is not None:
            return steps

        adapter_steps = self._adapter.get_size()
        if adapter_steps is not None:
            return adapter_steps

        raise ValueError(
            "When passing a generator, you " "must specify how many steps to draw."
        )

    @property
    def _samples(self):
        return self._adapter.get_samples()

    @property
    def batch_size(self):
        return self._adapter.batch_size


def _type_name(x):
    """Generates a description of the type of an object."""
    if isinstance(x, dict):
        key_types = set(_type_name(key) for key in x.keys())
        val_types = set(_type_name(key) for key in x.values())
        return "({} containing {} keys and {} values)".format(
            type(x), key_types, val_types
        )
    if isinstance(x, (list, tuple)):
        types = set(_type_name(val) for val in x)
        return "({} containing values of types {})".format(type(x), types)
    return str(type(x))


def select_data_adapter(x, y):
    """Selects a data adapter than can handle a given x and y."""
    adapter_cls = [cls for cls in ALL_ADAPTER_CLS if cls.can_handle(x, y)]
    if not adapter_cls:
        raise ValueError(
            "Failed to find data adapter that can handle "
            "input: {}, {}".format(_type_name(x), _type_name(y))
        )
    elif len(adapter_cls) > 1:
        raise RuntimeError(
            "Data adapters should be mutually exclusive for "
            "handling inputs. Found multiple adapters {} to handle "
            "input: {}, {}".format(adapter_cls, _type_name(x), _type_name(y))
        )
    return adapter_cls[0]



from unittest import TestCase

import elegy
import jax
import jax.numpy as jnp
import numpy as np


class TransformerTest(TestCase):
    def test_connects(self):
        transformer_model = elegy.nn.Transformer(
            head_size=64,
            num_heads=4,
            num_encoder_layers=2,
            num_decoder_layers=2,
        )

        src = np.random.uniform(size=(5, 32, 64))
        tgt = np.random.uniform(size=(5, 32, 64))

        _, params, collections = transformer_model.init(rng=elegy.RNGSeq(42))(src, tgt)
        out, params, collections = transformer_model.apply(
            params, collections, rng=elegy.RNGSeq(420)
        )(src, tgt)



import unittest
import elegy


import jax.numpy as jnp


class AccuracyTest(unittest.TestCase):
    #
    def test_basic(self):

        accuracy = elegy.metrics.Accuracy()

        result = accuracy.call_with_defaults()(
            y_true=jnp.array([1, 1, 1, 1]), y_pred=jnp.array([0, 1, 1, 1])
        )
        assert result == 0.75

        result = accuracy.call_with_defaults_jit()(
            jnp.array([1, 1, 1, 1]), jnp.array([1, 0, 0, 0])
        )
        assert result == 0.5



import typing as tp
from io import StringIO

import jax
import jax.numpy as jnp
import numpy as np
import yaml
from elegy import hooks, module, types, utils
from elegy.model import model_core
from elegy.generalized_module.generalized_module import (
    GeneralizedModule,
    generalize,
    is_generalizable,
)
from elegy.generalized_optimizer.generalized_optimizer import (
    GeneralizedOptimizer,
    generalize_optimizer,
)
from elegy.model.model_base import ModelBase
from elegy.optimizer import Optimizer
from tabulate import tabulate


class Model(ModelBase):
    """
    Model is a framework-agnostic trainer interface that is tasked with performing training, evaluation, and inference.
    It provides 2 main APIs:

    * The high-level API provides the Keras-like experience of simplicity and speed. The user provides things like the
        a modules, losses, metrics, and callbacks and Elegy takes care of the rest.
    * The low-level API provides the Pytorch Lightning-like experience of experisiveness and power. The user can
        override methods like `pred_step`, `test_step`, and `train_step` to perform jax-based computation with very
        few restrictions.

    ## High-level API
    To use the high-level API you first have to define a network architecture in a Module, currently we support Modules from
    Flax, Haiku, Elegy, and pure jax functions. Using `elegy.Module` this could look like this:

    ```python
    >>> import elegy, jax, optax
    >>> import jax.numpy as jnp

    >>> class MLP(elegy.Module):
    ...     def call(self, x: jnp.ndarray) -> jnp.ndarray:
    ...         x = elegy.nn.Linear(5)(x)
    ...         x = jax.nn.relu(x)
    ...         x = elegy.nn.Linear(2)(x)
    ...         return x

    ```


    You can pass this architecture to the Model API along with losses, metrics, optimizer, etc:
    ```python
    >>> model = elegy.Model(
    ...     module=MLP(),
    ...     loss=[
    ...         elegy.losses.SparseCategoricalCrossentropy(from_logits=True),
    ...         elegy.regularizers.GlobalL2(l=1e-5),
    ...     ],
    ...     metrics=elegy.metrics.SparseCategoricalAccuracy(),
    ...     optimizer=optax.rmsprop(1e-3),
    ... )

    ```

    Once the model is created, you can train the model with `model.fit()`, or use the model
    to do prediction with `model.predict()`.
    Checkout [Getting Started](https://poets-ai.github.io/elegy/getting-started) for
    additional details.

    ```python
    >>> x = jnp.ones(shape=[12, 10])
    >>> y = jnp.ones(shape=[12])

    >>> history = model.fit(x, y) # doctest: +SKIP

    ```

    Model supports optax optimizers as well as `elegy.Optimizer` which has a feature for definint + monitoring
    custom learning rate schedules, it is implemented on top of optax and you can use it like this:

    ```python
    >>> model = elegy.Model(
    ...     module=MLP(),
    ...     loss=elegy.losses.SparseCategoricalCrossentropy(from_logits=True),
    ...     metrics=elegy.metrics.SparseCategoricalAccuracy(),
    ...     optimizer=elegy.Optimizer(
    ...         # one or more optax optimizers as *args,
    ...         # these will be passed to optax.chain(...)
    ...         optax.adam(1.0), # <---- important to set this to 1.0
    ...
    ...         lr_schedule=lambda step, epoch: 1 / (epoch * 100 + step),
    ...         steps_per_epoch=1000,
    ...     ),
    ...     run_eagerly=True,
    ... )

    ```
    Notice how we set the learning rate parameter of the `adam` optimizer to `1.0`, this is necessary if you want the logged `lr`
    be closer to the "actual" learning rate since this feature was implemented by chaining an additional `optax.scale_by_schedule`
    at the end.

    ## High-level API

    """

    module: tp.Any = None
    loss: tp.Any = None
    metrics: tp.Any = None
    optimizer: tp.Any = None
    seed: int = 42

    api_module: tp.Optional[GeneralizedModule]
    api_loss: "Losses"
    api_metrics: "Metrics"
    api_optimizer: tp.Optional[GeneralizedOptimizer]

    def __init__(
        self,
        module: tp.Any = None,
        loss: tp.Any = None,
        metrics: tp.Any = None,
        optimizer: tp.Any = None,
        seed: int = 42,
        **kwargs,
    ):
        """[summary]

        Arguments:
            module: A `Module` instance.
            loss: A `elegy.Loss` or `Callable` instance representing the loss function of the network.
                You can define more loss terms by simply passing a possibly nested structure of
                lists and dictionaries of `elegy.Loss` or `Callable`s. Usually a plain list of losses is enough
                but using dictionaries will create namescopes for the names of the losses
                which might be useful e.g. to group things in tensorboard. Contrary to Keras convention,
                in Elegy there is no relation between the structure of `loss` with the structure
                of the labels and outputs of the network. Elegy's loss system is more flexible than
                the one provided by Keras, for more information on how to mimick Keras behavior checkout the
                [Losses and Metrics Guide](https://poets-ai.github.io/elegy/guides/losses-and-metrics)`.
            metrics: A `elegy.Metric` or `Callable` instance representing the loss function of the network.
                You can define more metrics terms by simply passing a possibly nested structure of
                lists and dictionaries of `elegy.Metric` or `Callable`s. Usually a plain list of metrics is enough
                but using dictionaries will create namescopes for the names of the metrics
                which might be useful e.g. to group things in tensorboard. Contrary to Keras convention,
                in Elegy there is no relation between the structure of `metrics` with the structure
                of the labels and outputs of the network. Elegy's metrics system is more flexible than
                the one provided by Keras, for more information on how to mimick Keras behavior checkout the
                [Losses and Metrics Guide](https://poets-ai.github.io/elegy/guides/losses-and-metrics)`.
            optimizer: A `optax` optimizer instance. Optix is a very flexible library for defining
                optimization pipelines with things like learning rate schedules, this means that
                there is no need for a `LearningRateScheduler` callback in Elegy.
            run_eagerly: Settable attribute indicating whether the model should run eagerly.
                Running eagerly means that your model will be run step by step, like Python code, instead of
                using Jax's `jit` to. Your model might run slower, but it should become easier for you to debug
                it by stepping into individual layer calls.
        """
        if "rng" in kwargs and not isinstance(kwargs["rng"], (int, types.RNGSeq)):
            raise ValueError(
                f"rng must be one of the following types: int, types.RNGSeq. Got {kwargs['rng']}"
            )

        super().__init__(**kwargs)

        # maybe add rng if initialized
        if self.initialized and (
            not hasattr(self.states, "rng") or self.states.rng is None
        ):
            self.states = self.states.update(rng=types.RNGSeq(seed))

        self.module = module
        self.loss = loss
        self.metrics = metrics
        self.optimizer = optimizer

        if loss is None:
            loss = {}

        if metrics is None:
            metrics = {}

        self.api_module = generalize(module) if module is not None else None
        self.api_loss = Losses(loss)
        self.api_metrics = Metrics(metrics)
        self.api_optimizer = (
            generalize_optimizer(optimizer) if optimizer is not None else None
        )
        self.seed = seed

    def __call__(self, *args, **kwargs):
        assert isinstance(self.states.rng, types.RNGSeq)
        assert self.module is not None

        return self.module.apply(
            self.states.net_params,
            self.states.net_states,
            self.states.rng,
        )(*args, **kwargs)

    def update_modules(self):
        if self.api_module is not None:
            self.api_module.update(
                params=self.states.net_params,
                states=self.states.net_states,
            )

    # ----------------------------------------------------------------
    # implement low-level API methods
    # ----------------------------------------------------------------
    def states_step(self) -> types.States:
        states = super().states_step()
        return states.update(
            rng=types.RNGSeq(self.seed),
            net_params=None,
            net_states=None,
            metrics_states=None,
            optimizer_states=None,
        )

    def init_step(
        self,
        x: tp.Any,
        y_true: tp.Any,
        sample_weight: tp.Optional[np.ndarray],
        class_weight: tp.Optional[np.ndarray],
        states: types.States,
    ) -> types.States:

        states = states.maybe_update(**self.states_step())
        training = True
        initializing = True

        _, states = utils.inject_dependencies(self.train_step)(
            x=x,
            y_true=y_true,
            sample_weight=sample_weight,
            class_weight=class_weight,
            states=states,
            initializing=initializing,
            training=training,
        )

        return states

    def summary_step(
        self,
        x: tp.Any,
        states: types.States,
    ) -> tp.List[types.SummaryTableEntry]:
        initializing = True
        training = True

        states = states.maybe_update(**self.states_step())

        with hooks.context(summaries=True):
            _, states = utils.inject_dependencies(self.pred_step)(
                x=x,
                states=states,
                initializing=initializing,
                training=training,
            )
            summaries = hooks.get_summaries()

        entries: tp.List[types.SummaryTableEntry] = []

        for path, module, value in summaries:

            module_params, module_states = self.api_module.get_summary_params(
                path=path,
                module=module,
                value=value,
                net_params=states.net_params,
                net_states=states.net_states,
            )

            entries.append(
                types.SummaryTableEntry(
                    path=("/".join(map(str, path)) if path else "*"),
                    module_type_name=(
                        module.__class__.__name__ if is_generalizable(module) else ""
                    ),
                    output_value=value,
                    trainable_params_count=(
                        utils.parameters_count(module_params)
                        if module_params is not None
                        else 0
                    ),
                    trainable_params_size=(
                        utils.parameters_bytes(module_params)
                        if module_params is not None
                        else 0
                    ),
                    non_trainable_params_count=(
                        utils.parameters_count(module_states)
                        if module_states is not None
                        else 0
                    ),
                    non_trainable_params_size=(
                        utils.parameters_bytes(module_states)
                        if module_states is not None
                        else 0
                    ),
                )
            )

        entries.append(
            types.SummaryTableEntry.totals_entry(
                trainable_params_count=utils.parameters_count(states.net_params),
                trainable_params_size=utils.parameters_bytes(states.net_params),
                non_trainable_params_count=utils.parameters_count(states.net_states),
                non_trainable_params_size=utils.parameters_bytes(states.net_states),
            )
        )

        return entries

    def pred_step(
        self,
        x: tp.Any,
        states: types.States,
        initializing: bool,
        training: bool,
    ) -> model_core.PredStep:

        if self.module is None:
            raise types.MissingModule(
                "Trying run default `pred_step` on a Model with no `module`, try overriding `pred_step`."
            )

        # [DI]
        x_args, x_kwargs = utils.get_input_args(
            x,
            states=states,
            initializing=initializing,
            training=training,
        )

        assert isinstance(states.rng, types.RNGSeq)

        if initializing:
            module_fn = self.api_module.init(states.rng)
        else:
            module_fn = self.api_module.apply(
                params=states.net_params,
                states=states.net_states,
                training=training,
                rng=states.rng,
            )

        y_pred, net_params, net_states = module_fn(*x_args, **x_kwargs)

        return model_core.PredStep(
            y_pred=y_pred,
            states=states.update(net_states=net_states, net_params=net_params),
        )

    def test_step(
        self,
        x: tp.Any,
        y_true: tp.Any,
        sample_weight: tp.Optional[np.ndarray],
        class_weight: tp.Optional[np.ndarray],
        states: types.States,
        initializing: bool,
        training: bool,
    ) -> model_core.TestStep:

        with hooks.context(losses=True, summaries=True):
            y_pred, states = utils.inject_dependencies(self.pred_step)(
                x=x,
                states=states,
                initializing=initializing,
                training=training,
            )

            aux_losses = hooks.get_losses()
            aux_metrics = hooks.get_metrics()

        assert isinstance(states.rng, types.RNGSeq)

        if initializing:
            metrics_fn = self.api_metrics.init(aux_metrics, states.rng)
            losses_fn = self.api_loss.init(aux_losses, states.rng)
        else:
            metrics_states, loss_states = states.metrics_states
            metrics_fn = self.api_metrics.apply(
                aux_metrics,
                states.rng,
                metrics_states,
            )
            losses_fn = self.api_loss.apply(aux_losses, loss_states)

        # [DI]metrics_states
        metrics_logs, metrics_states = metrics_fn(
            x=x,
            y_true=y_true,
            y_pred=y_pred,
            sample_weight=sample_weight,
            class_weight=class_weight,
            initializing=initializing,
            states=states,
            training=training,
        )

        # [DI]
        loss, loss_logs, loss_states = losses_fn(
            x=x,
            y_true=y_true,
            y_pred=y_pred,
            sample_weight=sample_weight,
            class_weight=class_weight,
            initializing=initializing,
            states=states,
            training=training,
        )

        logs = utils.merge_with_unique_names(metrics_logs, loss_logs)
        states = states.update(metrics_states=(metrics_states, loss_states))

        return model_core.TestStep(loss, logs, states)

    def grad_step(
        self,
        x: tp.Any,
        y_true: tp.Any,
        sample_weight: tp.Optional[np.ndarray],
        class_weight: tp.Optional[np.ndarray],
        states: types.States,
        initializing: bool,
        training: bool,
    ) -> model_core.GradStep:
        def loss_fn(
            net_params: tp.Any,
            states: types.States,
            x: tp.Any,
            y_true: tp.Any,
            sample_weight: tp.Optional[np.ndarray],
            class_weight: tp.Optional[np.ndarray],
        ):
            states = states.update(net_params=net_params)
            loss, logs, states = utils.inject_dependencies(self.test_step)(
                x=x,
                y_true=y_true,
                states=states,
                sample_weight=sample_weight,
                class_weight=class_weight,
                initializing=initializing,
                training=training,
            )

            return loss, (logs, states)

        (loss, (logs, states)), grads = jax.value_and_grad(loss_fn, has_aux=True)(
            states.net_params,
            states,
            x,
            y_true,
            sample_weight,
            class_weight,
        )

        return model_core.GradStep(loss, logs, states, grads)

    def train_step(
        self,
        x: tp.Any,
        y_true: tp.Any,
        sample_weight: tp.Optional[np.ndarray],
        class_weight: tp.Optional[np.ndarray],
        states: types.States,
        initializing: bool,
        training: bool,
    ) -> model_core.TrainStep:

        if initializing:
            loss, logs, states = utils.inject_dependencies(self.test_step)(
                x=x,
                y_true=y_true,
                states=states,
                sample_weight=sample_weight,
                class_weight=class_weight,
                initializing=initializing,
                training=training,
            )
            grads = None

        else:
            loss, logs, states, grads = utils.inject_dependencies(self.grad_step)(
                x=x,
                y_true=y_true,
                states=states,
                sample_weight=sample_weight,
                class_weight=class_weight,
                initializing=initializing,
                training=training,
            )

        if initializing and self.optimizer is None:
            return model_core.TrainStep(logs, states)
        elif self.optimizer is None:
            raise types.MissingOptimizer(
                "Trying to run `train_step` without an optimizer, "
                "please provide an optimizer to the Model(...) constructor or "
                "override `train_step`."
            )
        assert isinstance(states.rng, types.RNGSeq)

        # calculate current lr before update
        if initializing:
            optimizer_states = self.api_optimizer.init(states.rng, states.net_params)
            net_params = states.net_params
        else:
            assert grads is not None

            # TODO: add current_lr this to GeneralizedOptimizer and do this generically
            if isinstance(self.optimizer, Optimizer):
                lr = self.optimizer.current_lr(self.states.optimizer_states)

                if lr is not None:
                    logs["lr"] = lr

            net_params, optimizer_states = self.api_optimizer.apply(
                states.net_params,
                grads,
                states.optimizer_states,
                states.rng,
            )

        states = states.update(
            net_params=net_params,
            optimizer_states=optimizer_states,
        )

        return model_core.TrainStep(logs, states)

    # ----------------------------------------------------------------
    # Model-only methods
    # ----------------------------------------------------------------


class Metrics:
    metrics: tp.Dict[str, GeneralizedModule]

    def __init__(self, modules: tp.Any):
        names: tp.Set[str] = set()

        def get_name(module, path):
            name = utils.get_name(module)
            return f"{path}/{name}" if path else name

        self.metrics = {
            utils.get_unique_name(names, get_name(module, path)): generalize(
                module,
                callable_default=AvgMetric,
            )
            for path, module in utils.flatten_names(modules)
        }

    def calculate_metrics(
        self,
        aux_metrics: types.Logs,
        callback: tp.Callable[[str, GeneralizedModule], types.OutputStates],
    ) -> tp.Tuple[types.Logs, tp.Any]:

        states = {}

        for name, module in self.metrics.items():
            y_pred, _, states[name] = callback(name, module)

            names = set()
            for inner_name, inner_value in utils.flatten_names(y_pred):
                inner_name = f"{name}/{inner_name}" if inner_name else name
                inner_name = utils.get_unique_name(names, inner_name)

                aux_metrics[inner_name] = inner_value

        return aux_metrics, states

    def init(
        self,
        aux_metrics: types.Logs,
        rng: types.RNGSeq,
    ) -> tp.Callable[..., tp.Tuple[types.Logs, tp.Any]]:
        def lambda_(*args, **kwargs):
            def callback(name, module):
                kwargs_ = kwargs.copy()
                kwargs_["metrics_states"] = None
                return module.init(rng)(*args, **kwargs_)

            return self.calculate_metrics(aux_metrics, callback)

        return lambda_

    def apply(
        self,
        aux_metrics: types.Logs,
        rng: types.RNGSeq,
        states: tp.Any,
    ) -> tp.Callable[..., tp.Tuple[types.Logs, tp.Any]]:
        assert states is not None

        def lambda_(*args, **kwargs):
            def callback(name, module: GeneralizedModule):
                kwargs_ = kwargs.copy()
                kwargs_["metrics_states"] = states[name]

                return module.apply(
                    params=None, states=states[name], training=True, rng=rng
                )(*args, **kwargs_)

            return self.calculate_metrics(aux_metrics, callback)

        return lambda_


class AvgMetric(GeneralizedModule):
    def __init__(self, f: tp.Callable):
        self.f = f

    def init(self, rng: types.RNGSeq) -> tp.Callable[..., types.OutputStates]:
        def _lambda(*args, **kwargs) -> types.OutputStates:

            preds = utils.inject_dependencies(self.f)(*args, **kwargs)

            if isinstance(preds, types.OutputStates):
                return preds

            n = 0
            total = jax.tree_map(lambda x: jnp.zeros_like(x), preds)
            return types.OutputStates(
                preds=preds,
                params=None,
                states=(n, total),
            )

        return _lambda

    def apply(
        self,
        params: tp.Any,
        states: tp.Any,
        training: bool,
        rng: types.RNGSeq,
    ) -> tp.Callable[..., types.OutputStates]:
        def _lambda(*args, **kwargs) -> types.OutputStates:

            preds = utils.inject_dependencies(self.f)(*args, **kwargs)

            if isinstance(preds, types.OutputStates):
                return preds

            n, total = states
            n += 1
            total = jax.tree_multimap(lambda a, b: a + b, preds, total)
            preds = jax.tree_map(lambda total: total / n, total)
            return types.OutputStates(
                preds=preds,
                params=None,
                states=(n, total),
            )

        return _lambda


class Losses:
    losses: tp.Dict[str, tp.Callable]
    loss_metrics: "LossMetrics"

    def __init__(self, losses: tp.Any):
        names: tp.Set[str] = set()

        def get_name(loss_fn, path):
            name = utils.get_name(loss_fn)
            return f"{path}/{name}" if path else name

        self.losses = {
            utils.get_unique_name(names, get_name(loss_fn, path)): loss_fn
            for path, loss_fn in utils.flatten_names(losses)
        }
        self.loss_metrics = LossMetrics()

    def calculate_losses(self, *args, **kwargs) -> types.Logs:
        logs: types.Logs = {}

        for name, loss_fn in self.losses.items():
            losses = utils.inject_dependencies(loss_fn)(*args, **kwargs)

            names = set()
            for inner_name, loss in utils.flatten_names(losses):
                inner_name = f"{name}/{inner_name}" if inner_name else name
                inner_name = utils.get_unique_name(names, inner_name)

                logs[inner_name] = loss

        return logs

    def init(
        self,
        aux_losses: types.Logs,
        rng: types.RNGSeq,
    ) -> tp.Callable[..., tp.Tuple[types.Scalar, types.Logs, tp.Any]]:
        def _lambda(*args, **kwargs):
            module_logs = self.calculate_losses(*args, **kwargs)

            loss = sum(aux_losses.values(), 0.0) + sum(module_logs.values(), 0.0)
            loss_logs = dict(loss=loss)

            logs = utils.merge_with_unique_names(loss_logs, aux_losses, module_logs)

            names = set()
            logs = {
                utils.get_unique_name(names, f"{name}_loss")
                if "loss" not in name
                else utils.get_unique_name(names, name): value
                for name, value in logs.items()
            }

            logs, _, states = self.loss_metrics.init(rng=rng)(logs)

            return loss, logs, states

        return _lambda

    def apply(
        self,
        aux_losses: types.Logs,
        states: tp.Any,
    ) -> tp.Callable[..., tp.Tuple[types.Scalar, types.Logs, tp.Any]]:
        def _lambda(*args, **kwargs):
            module_logs = self.calculate_losses(*args, **kwargs)

            loss = sum(aux_losses.values(), 0.0) + sum(module_logs.values(), 0.0)
            loss_logs = dict(loss=loss)

            logs = utils.merge_with_unique_names(loss_logs, aux_losses, module_logs)

            names = set()
            logs = {
                utils.get_unique_name(names, f"{name}_loss")
                if "loss" not in name
                else utils.get_unique_name(names, name): value
                for name, value in logs.items()
            }

            parameters = None
            logs, _, states_ = self.loss_metrics.apply(parameters, states)(logs)

            return loss, logs, states_

        return _lambda


class LossMetrics(module.Module):
    def call(self, logs):

        count = self.add_parameter(
            "count", lambda: jnp.array(0, dtype=jnp.int32), trainable=False
        )
        total = self.add_parameter(
            "total", lambda: jax.tree_map(jnp.zeros_like, logs), trainable=False
        )

        count = count + 1
        total = jax.tree_multimap(lambda a, b: a + b, total, logs)

        self.update_parameter("count", count)
        self.update_parameter("total", total)

        return jax.tree_map(lambda total: total / count, total)



from .callback_list import CallbackList
from .callback import Callback
from .history import History
from .model_checkpoint import ModelCheckpoint
from .early_stopping import EarlyStopping
from .lambda_callback import LambdaCallback
from .terminate_nan import TerminateOnNaN
from .remote_monitor import RemoteMonitor
from .csv_logger import CSVLogger
from .tensorboard import TensorBoard

__all__ = [
    "CallbackList",
    "Callback",
    "History",
    "ModelCheckpoint",
    "EarlyStopping",
    "LambdaCallback",
    "TerminateOnNaN",
    "RemoteMonitor",
    "CSVLogger",
    "TensorBoard",
]



from elegy import types
from elegy import utils
import typing as tp

import jax.numpy as jnp

from elegy.metrics.mean import Mean
from elegy.metrics.reduce_confusion_matrix import Reduction, ReduceConfusionMatrix
from elegy.metrics.metric import Metric
from elegy.metrics import Precision
from elegy.metrics import Recall


def f1(
    y_true: jnp.ndarray,
    y_pred: jnp.ndarray,
    sample_weight: jnp.ndarray,
    precision: Precision,
    recall: Recall,
) -> jnp.ndarray:

    precision_values = precision(
        y_true=y_true, y_pred=y_pred, sample_weight=sample_weight
    )

    recall_values = recall(y_true=y_true, y_pred=y_pred, sample_weight=sample_weight)

    return 2 * jnp.divide(
        (precision_values * recall_values), (precision_values + recall_values)
    )


class F1(Metric):
    """
    The metric calculates the Harmonic mean between precision and recall. This value is ultimately returned as
    `f1`.

    If `sample_weight` is `None`, weights default to 1. Use `sample_weight` of 0 to mask values.

    If sample_weight is None, weights default to 1. Use sample_weight of 0 to mask values.

    If class_id is specified, we calculate precision by considering only the entries in the batch
    for which class_id is above the threshold and computing the fraction of them for which class_id
    is indeed a correct label.

    ```python
        f1 = elegy.metrics.F1()

        result = f1(
            y_true=jnp.array([0, 1, 1, 1]), y_pred=jnp.array([1, 0, 1, 1])
        )
        assert result == 0.6666667 # 2 * (0.44445 / 1.33334)

        result = f1(
            y_true=jnp.array([1, 1, 1, 1]), y_pred=jnp.array([1, 1, 0, 0])
        )
        assert result == 666667 # 2 * (0.5 / 1.5)
    ```

    Usage with elegy API:

    ```python
        model = elegy.Model(
        module_fn,
        loss=elegy.losses.CategoricalCrossentropy(),
        metrics=elegy.metrics.F1(),
        optimizer=optix.adam(1e-3),
    )
    ```
    """

    def __init__(
        self,
        on: tp.Optional[types.IndexLike] = None,
        threshold=None,
        class_id=None,
        **kwargs
    ):
        """
        Creates a `F1` instance.

        Arguments:
            on: A string or integer, or iterable of string or integers, that
                indicate how to index/filter the `y_true` and `y_pred`
                arguments before passing them to `call`. For example if `on = "a"` then
                `y_true = y_true["a"]`. If `on` is an iterable
                the structures will be indexed iteratively, for example if `on = ["a", 0, "b"]`
                then `y_true = y_true["a"][0]["b"]`, same for `y_pred`. For more information
                check out [Keras-like behavior](https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#keras-like-behavior).

            threshold: (Optional) A float value or a python list/tuple of float threshold
                values in [0, 1]. A threshold is compared with prediction values to determine
                the truth value of predictions (i.e., above the threshold is true, below is false).
                One metric value is generated for each threshold value. If neither threshold is set
                the default is to calculate precision with threshold=0.5.

            class_id: (Optional) Integer class ID for which we want binary metrics.
                This must be in the half-open interval `[0, num_classes)`, where
                `num_classes` is the last dimension of predictions.

            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(on=on, **kwargs)
        self.threshold = 0.5 if threshold is None else threshold
        self.class_id = 1 if class_id is None else class_id
        self.precision = Precision(threshold=self.threshold)
        self.recall = Recall(threshold=self.threshold)

    def call(
        self,
        y_true: jnp.ndarray,
        y_pred: jnp.ndarray,
        sample_weight: tp.Optional[jnp.ndarray] = None,
    ) -> jnp.ndarray:
        """
        Accumulates f1 values (armonic mean between precision and recall).

        Arguments:
            y_true: Ground truth values. shape = `[batch_size, d0, .. dN]`.

            y_pred: The predicted values. shape = `[batch_size, d0, .. dN]`.

            sample_weight: Optional weighting of each example. Defaults to 1. Can be a
                `Tensor` whose rank is either 0, or the same rank as `y_true`, and must
                be broadcastable to `y_true`.
        Returns:
            Array with the cumulative f1.
        """

        return f1(
            y_true=y_true,
            y_pred=y_pred,
            sample_weight=sample_weight,
            precision=self.precision,
            recall=self.recall,
        )



# Lint as: python3
# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Haiku initializers."""

from elegy import module
from elegy import hooks
from elegy import types
import jax
import jax.numpy as jnp
import numpy as np

__all__ = [
    "Constant",
    "Orthogonal",
    "RandomNormal",
    "RandomUniform",
    "TruncatedNormal",
    "UniformScaling",
    "VarianceScaling",
]


def _compute_fans(shape):
    """Computes the number of input and output units for a weight shape."""
    if len(shape) < 1:
        fan_in = fan_out = 1
    elif len(shape) == 1:
        fan_in = fan_out = shape[0]
    elif len(shape) == 2:
        fan_in, fan_out = shape
    else:
        # Assuming convolution kernels (2D, 3D, or more.)
        # kernel_shape: (..., input_depth, depth)
        receptive_field_size = np.prod(shape[:-2])
        fan_in = shape[-2] * receptive_field_size
        fan_out = shape[-1] * receptive_field_size
    return fan_in, fan_out


class Constant(types.Initializer):
    """Initializes with a constant."""

    def __init__(self, constant):
        """
        Constructs a Constant initializer.

        Args:
            constant: Constant to initialize with.
        """
        self.constant = constant

    def __call__(self, shape: types.Shape, dtype: types.DType) -> np.ndarray:
        return jnp.broadcast_to(self.constant, shape).astype(dtype)


class RandomNormal(types.Initializer):
    """Initializes by sampling from a normal distribution."""

    def __init__(self, stddev=1.0, mean=0.0):
        """
        Constructs a RandomNormal initializer.

        Args:
            stddev: The standard deviation of the normal distribution to sample from.
            mean: The mean of the normal distribution to sample from.
        """
        self.stddev = stddev
        self.mean = mean

    def __call__(self, shape: types.Shape, dtype: types.DType) -> np.ndarray:
        m = jax.lax.convert_element_type(self.mean, dtype)
        s = jax.lax.convert_element_type(self.stddev, dtype)
        return m + s * jax.random.normal(module.next_key(), shape, dtype)


class TruncatedNormal(types.Initializer):
    """Initializes by sampling from a truncated normal distribution."""

    def __init__(self, stddev=1.0, mean=0.0):
        """
        Constructs a RandomNormal initializer.

        Args:
            stddev: The standard deviation parameter of the truncated
                normal distribution.
            mean: The mean of the truncated normal distribution.
        """
        self.stddev = stddev
        self.mean = mean

    def __call__(self, shape: types.Shape, dtype: types.DType) -> np.ndarray:
        m = jax.lax.convert_element_type(self.mean, dtype)
        s = jax.lax.convert_element_type(self.stddev, dtype)
        unscaled = jax.random.truncated_normal(
            module.next_key(), -2.0, 2.0, shape, dtype
        )
        return s * unscaled + m


class RandomUniform(types.Initializer):
    """Initializes by sampling from a uniform distribution."""

    def __init__(self, minval=0.0, maxval=1.0):
        """
        Constructs a RandomUniform initializer.

        Args:
            minval: The lower limit of the uniform distribution.
            maxval: The upper limit of the uniform distribution.
        """
        self.minval = minval
        self.maxval = maxval

    def __call__(self, shape: types.Shape, dtype: types.DType) -> np.ndarray:
        return jax.random.uniform(
            module.next_key(), shape, dtype, self.minval, self.maxval
        )


class VarianceScaling(types.Initializer):
    """types.Initializer which adapts its scale to the shape of the initialized array.

    The initializer first computes the scaling factor ``s = scale / n``, where n
    is:

      - Number of input units in the weight tensor, if ``mode = fan_in``.
      - Number of output units, if ``mode = fan_out``.
      - Average of the numbers of input and output units, if ``mode = fan_avg``.

    Then, with ``distribution="truncated_normal" or "normal"``,
    samples are drawn from a distribution with a mean of zero and a standard
    deviation (after truncation, if used) ``stddev = sqrt(s)``.

    With ``distribution=uniform``, samples are drawn from a uniform distribution
    within ``[-limit, limit]``, with ``limit = sqrt(3 * s)``.

    The variance scaling initializer can be configured to generate other standard
    initializers using the scale, mode and distribution arguments. Here are some
    example configurations:

    ==============  ==============================================================
    Name            Parameters
    ==============  ==============================================================
    glorot_uniform  scale=1.0, mode=``fan_avg``, distribution=``uniform``
    glorot_normal   scale=1.0, mode=``fan_avg``, distribution=``truncated_normal``
    lecun_uniform   scale=1.0, mode=``fan_in``,  distribution=``uniform``
    lecun_normal    scale=1.0, mode=``fan_in``,  distribution=``truncated_normal``
    he_uniform      scale=2.0, mode=``fan_in``,  distribution=``uniform``
    he_normal       scale=2.0, mode=``fan_in``,  distribution=``truncated_normal``
    ==============  ==============================================================
    """

    def __init__(self, scale=1.0, mode="fan_in", distribution="truncated_normal"):
        """
        Constructs the VarianceScaling initializer.

        Args:
            scale: Scale to multiply the variance by.
            mode: One of ``fan_in``, ``fan_out``, ``fan_avg``
            distribution: Random distribution to use. One of ``truncated_normal``,
                ``normal`` or ``uniform``.
        """
        if scale <= 0.0:
            raise ValueError("`scale` must be a positive float.")
        if mode not in {"fan_in", "fan_out", "fan_avg"}:
            raise ValueError("Invalid `mode` argument:", mode)
        distribution = distribution.lower()
        if distribution not in {"normal", "truncated_normal", "uniform"}:
            raise ValueError("Invalid `distribution` argument:", distribution)
        self.scale = scale
        self.mode = mode
        self.distribution = distribution

    def __call__(self, shape: types.Shape, dtype: types.DType) -> np.ndarray:
        scale = self.scale
        fan_in, fan_out = _compute_fans(shape)
        if self.mode == "fan_in":
            scale /= max(1.0, fan_in)
        elif self.mode == "fan_out":
            scale /= max(1.0, fan_out)
        else:
            scale /= max(1.0, (fan_in + fan_out) / 2.0)

        if self.distribution == "truncated_normal":
            stddev = np.sqrt(scale)
            # Adjust stddev for truncation.
            # Constant from scipy.stats.truncnorm.std(a=-2, b=2, loc=0., scale=1.)
            distribution_stddev = np.asarray(0.87962566103423978, dtype=dtype)
            stddev = stddev / distribution_stddev
            return TruncatedNormal(stddev=stddev)(shape, dtype)
        elif self.distribution == "normal":
            stddev = np.sqrt(scale)
            return RandomNormal(stddev=stddev)(shape, dtype)
        else:
            limit = np.sqrt(3.0 * scale)
            return RandomUniform(minval=-limit, maxval=limit)(shape, dtype)


class UniformScaling(types.Initializer):
    """
    Uniform scaling initializer.

    Initializes by sampling from a uniform distribution, but with the variance
    scaled by the inverse square root of the number of input units, multiplied by
    the scale.
    """

    def __init__(self, scale=1.0):
        """Constructs the UniformScaling initializer.

        Args:
          scale: Scale to multiply the upper limit of the uniform distribution by.
        """
        self.scale = scale

    def __call__(self, shape: types.Shape, dtype: types.DType) -> np.ndarray:
        input_size = np.product(shape[:-1])
        max_val = np.sqrt(3 / input_size) * self.scale
        return RandomUniform(-max_val, max_val)(shape, dtype)


class Orthogonal(types.Initializer):
    """Uniform scaling initializer."""

    def __init__(self, scale=1.0, axis=-1):
        """
        Construct an initializer for uniformly distributed orthogonal matrices.

        These matrices will be row-orthonormal along the access specified by `axis`.
        If the rank of the weight is greater than 2, the shape will be flattened in
        all other dimensions and then will be row-orthonormal along the final
        dimension. Note that this only works if the `axis` dimension is larger,
        otherwise the matrix will be transposed (equivalently, it will be column
        orthonormal instead of row orthonormal).

        Args:
            scale: Scale factor.
            axis (int): Which axis corresponds to the "output dimension" of the
                tensor.


        If the shape is not square, the matrices will have orthonormal rows or
        columns depending on which side is smaller.
        """
        self.scale = scale
        self.axis = axis

    def __call__(self, shape: types.Shape, dtype: types.DType) -> np.ndarray:
        if len(shape) < 2:
            raise ValueError("Orthogonal initializer requires at least a 2D shape.")
        n_rows = shape[self.axis]
        n_cols = np.prod(shape) // n_rows
        matrix_shape = (n_rows, n_cols) if n_rows > n_cols else (n_cols, n_rows)
        norm_dst = jax.random.normal(module.next_key(), matrix_shape, dtype)
        q_mat, r_mat = jnp.linalg.qr(norm_dst)
        # Enforce Q is uniformly distributed
        q_mat *= jnp.sign(jnp.diag(r_mat))
        if n_rows < n_cols:
            q_mat = q_mat.T
        q_mat = jnp.reshape(q_mat, (n_rows,) + tuple(np.delete(shape, self.axis)))
        q_mat = jnp.moveaxis(q_mat, 0, self.axis)
        return jax.lax.convert_element_type(self.scale, dtype) * q_mat



from .batch_normalization import BatchNormalization
from .conv import Conv1D, Conv2D, Conv3D, ConvND
from .dropout import Dropout
from .embedding import Embedding, EmbedLookupStyle
from .flatten import Flatten, Reshape
from .layer_normalization import InstanceNormalization, LayerNormalization
from .linear import Linear
from .moving_averages import EMAParamsTree
from .multi_head_attention import MultiHeadAttention
from .pool import AvgPool, MaxPool
from .sequential_module import Sequential, sequential
from .transformers import (
    Transformer,
    TransformerDecoder,
    TransformerDecoderLayer,
    TransformerEncoder,
    TransformerEncoderLayer,
)

__all__ = [
    "EMAParamsTree",
    "BatchNormalization",
    "MultiHeadAttention",
    "Transformer",
    "TransformerDecoder",
    "TransformerDecoderLayer",
    "TransformerEncoder",
    "TransformerEncoderLayer",
    "Conv1D",
    "Conv2D",
    "Conv3D",
    "ConvND",
    "Dropout",
    "Flatten",
    "Reshape",
    "Linear",
    "Sequential",
    "sequential",
    "LayerNormalization",
    "InstanceNormalization",
    "Embedding",
    "EmbedLookupStyle",
    "MaxPool",
    "AvgPool",
]



import elegy


from elegy import utils, types
import jax.numpy as jnp
import jax
import tensorflow.keras as tfk

# import debugpy

# print("Waiting for debugger...")
# debugpy.listen(5679)
# debugpy.wait_for_client()


def test_basic():

    y_true = jnp.array([[0.0, 1.0], [1.0, 1.0]])
    y_pred = jnp.array([[1.0, 0.0], [1.0, 1.0]])

    # Using 'auto'/'sum_over_batch_size' reduction type.
    cosine_loss = elegy.losses.CosineSimilarity(axis=1)
    assert cosine_loss(y_true, y_pred) == -0.49999997

    # Calling with 'sample_weight'.
    assert (
        cosine_loss(y_true, y_pred, sample_weight=jnp.array([0.8, 0.2])) == -0.099999994
    )

    # Using 'sum' reduction type.
    cosine_loss = elegy.losses.CosineSimilarity(
        axis=1, reduction=elegy.losses.Reduction.SUM
    )
    assert cosine_loss(y_true, y_pred) == -0.99999994

    # Using 'none' reduction type.
    cosine_loss = elegy.losses.CosineSimilarity(
        axis=1, reduction=elegy.losses.Reduction.NONE
    )

    assert jnp.equal(cosine_loss(y_true, y_pred), jnp.array([-0.0, -0.99999994])).all()


def test_function():

    rng = jax.random.PRNGKey(42)

    y_true = jax.random.randint(rng, shape=(2, 3), minval=0, maxval=2)
    y_pred = jax.random.uniform(rng, shape=(2, 3))

    loss = elegy.losses.cosine_similarity(y_true, y_pred, axis=1)
    assert loss.shape == (2,)

    y_true = y_true / jnp.maximum(
        jnp.linalg.norm(y_true, axis=1, keepdims=True), jnp.sqrt(types.EPSILON)
    )
    y_pred = y_pred / jnp.maximum(
        jnp.linalg.norm(y_pred, axis=1, keepdims=True), jnp.sqrt(types.EPSILON)
    )
    assert jnp.array_equal(loss, -jnp.sum(y_true * y_pred, axis=1))


def test_compatibility():
    # Input:  true (y_true) and predicted (y_pred) tensors
    rng = jax.random.PRNGKey(121)

    y_true = jax.random.randint(rng, shape=(2, 3), minval=0, maxval=2)
    y_true = y_true.astype(dtype=jnp.float32)
    y_pred = jax.random.uniform(rng, shape=(2, 3))

    # cosine_loss using sample_weight
    cosine_loss = elegy.losses.CosineSimilarity(axis=1)
    cosine_loss_tfk = tfk.losses.CosineSimilarity(axis=1)

    assert jnp.isclose(
        cosine_loss(y_true, y_pred, sample_weight=jnp.array([1, 0])),
        cosine_loss_tfk(y_true, y_pred, sample_weight=jnp.array([1, 0])),
        rtol=0.0001,
    )

    # cosine_loss with reduction method: SUM
    cosine_loss = elegy.losses.CosineSimilarity(
        axis=1, reduction=elegy.losses.Reduction.SUM
    )
    cosine_loss_tfk = tfk.losses.CosineSimilarity(
        axis=1, reduction=tfk.losses.Reduction.SUM
    )
    assert jnp.isclose(
        cosine_loss(y_true, y_pred), cosine_loss_tfk(y_true, y_pred), rtol=0.0001
    )

    # cosine_loss with reduction method: NONE
    cosine_loss = elegy.losses.CosineSimilarity(
        axis=1, reduction=elegy.losses.Reduction.NONE
    )
    cosine_loss_tfk = tfk.losses.CosineSimilarity(
        axis=1, reduction=tfk.losses.Reduction.NONE
    )
    assert jnp.all(
        jnp.isclose(
            cosine_loss(y_true, y_pred), cosine_loss_tfk(y_true, y_pred), rtol=0.0001
        )
    )


if __name__ == "__main__":

    test_basic()
    test_function()
    test_compatibility()



import unittest

import elegy
import jax.numpy as jnp
import numpy as np
import torch
from torch.utils.data import DataLoader, TensorDataset


class TestModelBase(unittest.TestCase):
    def test_predict(self):
        class Model(elegy.model.model_base.ModelBase):
            def init_step(self, x, states):
                _, states = self.pred_step(x, True, states)
                return states

            def pred_step(self, x, initializing, states):
                if initializing:
                    states = elegy.States(net_states=0)
                else:
                    states = elegy.States(net_states=states.net_states + 1)

                return elegy.PredStep(x + 1.0, states)

        model = Model()

        x = np.random.uniform(size=(100, 1))
        model.init(x, batch_size=50)
        y = model.predict(x, batch_size=50)

        assert np.allclose(y, x + 1.0)
        assert model.states.net_states == 2
        assert model.initial_states.net_states == 0

    def test_evaluate(self):
        class Model(elegy.model.model_base.ModelBase):
            def init_step(self, x, states):
                _, _, states = self.test_step(x, True, states)
                return states

            def test_step(self, x, initializing, states):
                if initializing:
                    states = elegy.States(metrics_states=0)
                else:
                    states = elegy.States(metrics_states=states.metrics_states + 1)

                return elegy.TestStep(
                    loss=0.1,
                    logs=dict(loss=jnp.sum(x)),
                    states=states,
                )

            def train_step(self):
                ...

        model = Model(run_eagerly=True)

        x = np.random.uniform(size=(100, 1))

        logs = model.evaluate(x, batch_size=100)
        assert np.allclose(logs["loss"], np.sum(x))
        assert model.states.metrics_states == 1

        logs = model.evaluate(x, batch_size=50)
        assert np.allclose(logs["loss"], np.sum(x[50:]))
        assert model.states.metrics_states == 2

    def test_fit(self):
        class Model(elegy.model.model_base.ModelBase):
            def init_step(self, x, states):
                _, states = self.train_step(x, states, True)
                return states

            def train_step(self, x, states, initializing):
                if initializing:
                    states = elegy.States(optimizer_states=0)
                else:
                    states = elegy.States(optimizer_states=states.optimizer_states + 1)

                return elegy.TrainStep(
                    logs=dict(loss=jnp.sum(x)),
                    states=states,
                )

        model = Model()

        x = np.random.uniform(size=(100, 1))

        history = model.fit(x, batch_size=100)
        assert np.allclose(history.history["loss"], np.sum(x))
        assert model.states.optimizer_states == 1

        history = model.fit(x, batch_size=50, shuffle=False)
        assert np.allclose(history.history["loss"], np.sum(x[50:]))
        assert model.states.optimizer_states == 3

    def test_init(self):
        class Model(elegy.model.model_base.ModelBase):
            def init_step(self, x, y_true, states: elegy.States):
                return states.update(a=x.shape, b=y_true.shape)

        model = Model()

        x = np.random.uniform(size=(10, 1))
        y = np.random.uniform(size=(10, 3))

        model.init(x=x, y=y, batch_size=2)

        assert model.states.a == (2, 1)
        assert model.states.b == (2, 3)
        assert model.initialized

    def test_init_dataloader(self):
        class Model(elegy.model.model_base.ModelBase):
            def init_step(self, x, y_true, states: elegy.States):
                return states.update(a=x.shape, b=y_true.shape)

            def pred_step(self, x, states):
                states = states.update(c=3)
                return x + 1, states

        model = Model()

        x = np.random.uniform(size=(10, 1))
        y = np.random.uniform(size=(10, 3))

        dataset = TensorDataset(torch.from_numpy(x), torch.from_numpy(y))
        dataloader = DataLoader(dataset, batch_size=2)

        model.init(x=dataloader, batch_size=2)

        assert model.states.a == (2, 1)
        assert model.states.b == (2, 3)
        assert model.initialized

        y_pred = model.predict(x=dataloader)
        assert jnp.allclose(y_pred, x + 1)
        y_pred = model.predict(x=dataloader)
        assert jnp.allclose(y_pred, x + 1)
        y_pred

    def test_init_predict(self):
        class Model(elegy.model.model_base.ModelBase):
            def init_step(self, x, states: elegy.States):
                return states.update(a=x.shape)

            def pred_step(self, x, states):
                states = states.update(c=3)
                return x + 1, states

        model = Model()

        x = np.random.uniform(size=(10, 1))

        model.init(x=x, batch_size=2)
        y_pred = model.predict(x=x, batch_size=2)

        assert jnp.allclose(y_pred, x + 1)
        assert model.states.a == (2, 1)
        assert model.states.c == 3
        assert model.initialized



from unittest import TestCase

import jax.numpy as jnp
import tensorflow.keras as tfk
import numpy as np
import elegy


class MAETest(TestCase):
    def test_basic(self):

        y_true = jnp.array([1, 1, 1, 1])
        y_pred = jnp.array([0, 1, 1, 1])

        assert np.allclose(
            elegy.metrics.MeanAbsoluteError().call_with_defaults()(y_true, y_pred),
            tfk.metrics.MeanAbsoluteError()(y_true, y_pred),
        )

        y_true = jnp.array([1, 1, 1, 1])
        y_pred = jnp.array([1, 0, 0, 0])

        assert np.allclose(
            elegy.metrics.MeanAbsoluteError().call_with_defaults()(y_true, y_pred),
            tfk.metrics.MeanAbsoluteError()(y_true, y_pred),
        )

        y_true = (np.random.uniform(0, 1, size=(5, 6, 7)) > 0.5).astype(np.int32)
        y_pred = np.random.uniform(0, 1, size=(5, 6, 7))

        assert np.allclose(
            tfk.metrics.MeanAbsoluteError()(y_true, y_pred),
            elegy.metrics.MeanAbsoluteError().call_with_defaults()(y_true, y_pred),
        )



# Implementation based on tf.keras.engine.data_adapter.py
# https://github.com/tensorflow/tensorflow/blob/2b96f3662bd776e277f86997659e61046b56c315/tensorflow/python/keras/engine/data_adapter.py


from tensorflow.python.data.experimental.ops import cardinality
from tensorflow.python.data.ops import dataset_ops

from .data_adapter import DataAdapter
from .utils import is_none_or_empty, map_structure, flatten


class TFDatasetAdapter(DataAdapter):
    """Adapter that handles `tf.data.Dataset`."""

    @staticmethod
    def can_handle(x, y=None):
        return isinstance(x, (dataset_ops.DatasetV1, dataset_ops.DatasetV2))

    def __init__(self, x, y=None, sample_weights=None, steps=None, **kwargs):
        super().__init__(x, y, **kwargs)
        # Note that the dataset instance is immutable, its fine to reuse the user
        # provided dataset.
        self._dataset = x

        # The user-provided steps.
        self._user_steps = steps

        self._validate_args(y, sample_weights, steps)

        # Since we have to know the dtype of the dataset when we build the
        # dataset, we have to look at a batch to infer the structure.
        peek = next(iter(x))

        self._first_batch_size = int(list(flatten(peek))[0].shape[0])

    def get_dataset(self):
        def parse_tf_data_gen():
            for batch in iter(self._dataset):
                batch = map_structure(lambda x: x.numpy(), batch)
                yield batch

        return parse_tf_data_gen

    def get_size(self):
        size = cardinality.cardinality(self._dataset)
        if size == cardinality.INFINITE and self._user_steps is None:
            raise ValueError(
                "When passing an infinitely repeating tf.data.Dataset, you "
                "must specify how many steps to draw."
            )
        elif size == cardinality.INFINITE:
            return self._user_steps
        elif size >= 0:
            return size.numpy().item()

    @property
    def batch_size(self):
        return self.representative_batch_size

    @property
    def representative_batch_size(self):
        return self._first_batch_size

    @property
    def partial_batch_size(self):
        return

    def has_partial_batch(self):
        return False

    def should_recreate_iterator(self):
        # If user doesn't supply `steps`, or if they supply `steps` that
        # exactly equals the size of the `Dataset`, create a new iterator
        # each epoch.
        return (
            self._user_steps is None
            or cardinality.cardinality(self._dataset).numpy() == self._user_steps
        )

    def _validate_args(self, y, sample_weights, steps):
        """Validates `__init__` arguments."""
        # Arguments that shouldn't be passed.
        if not is_none_or_empty(y):
            raise ValueError(
                "`y` argument is not supported when using " "tf.Data.dataset as input."
            )
        if not is_none_or_empty(sample_weights):
            raise ValueError(
                "`sample_weight` argument is not supported when using "
                "tf.Data.dataset as input."
            )

        size = cardinality.cardinality(self._dataset).numpy()
        if size == cardinality.INFINITE and steps is None:
            raise ValueError(
                "When providing an infinitely repeating tf.data.Dataset, you must specify "
                "the number of steps to run."
            )



import itertools
import typing as tp

from .data_adapter import DataAdapter
from .utils import (
    assert_not_namedtuple,
    flatten,
    is_none_or_empty,
    pack_x_y_sample_weight,
    unpack_x_y_sample_weight,
)


class GeneratorDataAdapter(DataAdapter):
    """Adapter that handles python generators and iterators."""

    @staticmethod
    def can_handle(x, y=None):
        return (hasattr(x, "__next__") or hasattr(x, "next")) and hasattr(x, "__iter__")

    def __init__(
        self,
        x: tp.Union[tp.Iterable],
        y=None,
        sample_weights=None,
        **kwargs,
    ):
        # Generators should never shuffle as exhausting the generator in order to
        # shuffle the batches is inefficient.
        kwargs.pop("shuffle", None)

        if not is_none_or_empty(y):
            raise ValueError(
                "`y` argument is not supported when using " "python generator as input."
            )
        if not is_none_or_empty(sample_weights):
            raise ValueError(
                "`sample_weight` argument is not supported when using "
                "python generator as input."
            )

        super(GeneratorDataAdapter, self).__init__(x, y, **kwargs)

        # Since we have to know the dtype of the python generator when we build the
        # dataset, we have to look at a batch to infer the structure.
        peek, x = self._peek_and_restore(x)
        assert_not_namedtuple(peek)
        peek = self._standardize_batch(peek)

        self._first_batch_size = int(list(flatten(peek))[0].shape[0])

        def wrapped_generator():
            for data in x:
                yield self._standardize_batch(data)

        dataset = wrapped_generator

        self._dataset = dataset

    def _standardize_batch(self, data):
        """Standardizes a batch output by a generator."""
        # Removes `None`s.
        x, y, sample_weight = unpack_x_y_sample_weight(data)
        data = pack_x_y_sample_weight(x, y, sample_weight)

        return data

    @staticmethod
    def _peek_and_restore(x):
        peek = next(x)
        return peek, itertools.chain([peek], x)

    def get_dataset(self):
        return self._dataset

    def get_size(self):
        return None

    @property
    def batch_size(self):
        return self.representative_batch_size

    @property
    def representative_batch_size(self):
        return self._first_batch_size

    def has_partial_batch(self):
        return False

    @property
    def partial_batch_size(self):
        return

    def should_recreate_iterator(self):
        return False



__version__ = "0.7.4"


from elegy import types
from elegy.utils import inject_dependencies

from . import losses  # module,
from . import (
    callbacks,
    hooks,
    initializers,
    metrics,
    model,
    nets,
    nn,
    regularizers,
    utils,
    data,
)
from .losses import Loss
from .metrics import Metric
from .model.model import Metrics, Model, Losses
from .model.model_core import PredStep, TestStep, GradStep, TrainStep
from .model.model_base import ModelBase, load
from .module import Module, to_module
from .optimizer import Optimizer
from .types import (
    RNGSeq,
    States,
    OutputStates,
)
from .generalized_module.generalized_module import GeneralizedModule
from .generalized_optimizer.generalized_optimizer import GeneralizedOptimizer

try:
    from .generalized_module.linen_module import flax_summarize, flax_summary
except types.DependencyUnavailable as e:
    pass
try:
    from .generalized_module.haiku_module import (
        HaikuModule,
        haiku_summarize,
        haiku_summary,
    )
except types.DependencyUnavailable as e:
    pass


__all__ = [
    "GeneralizedModule",
    "GeneralizedOptimizer",
    "GradStep",
    "HaikuModule",
    "Loss",
    "Losses",
    "Metric",
    "Metrics",
    "Model",
    "ModelBase",
    "Module",
    "Optimizer",
    "OutputStates",
    "PredStep",
    "RNGSeq",
    "States",
    "TestStep",
    "TrainStep",
    "callbacks",
    "data",
    "flax_summarize",
    "flax_summary",
    "hooks",
    "initializers",
    "inject_dependencies",
    "losses",
    "metrics",
    "nets",
    "nn",
    "regularizers",
    "to_module",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['elegy',
 'elegy.callbacks',
 'elegy.data',
 'elegy.generalized_module',
 'elegy.generalized_optimizer',
 'elegy.losses',
 'elegy.metrics',
 'elegy.model',
 'elegy.nets',
 'elegy.nn',
 'elegy.regularizers']

package_data = \
{'': ['*']}

install_requires = \
['cloudpickle>=1.5.0,<2.0.0',
 'dm-haiku>=0.0.2,<0.0.3',
 'jax>=0.2.13,<0.3.0',
 'jaxlib>=0.1.67,<0.2.0',
 'numpy>=1.0.0,<2.0.0',
 'optax>=0.0.6,<0.0.7',
 'pyyaml>=5.3.1,<6.0.0',
 'rich>=9.11.1,<10.0.0',
 'tabulate>=0.8.7,<0.9.0',
 'tensorboardx>=2.1,<3.0',
 'toolz>=0.10.0,<0.11.0']

extras_require = \
{':python_version < "3.8"': ['typing_extensions>=3.7.4,<4.0.0'],
 ':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.7,<0.8']}

setup_kwargs = {
    'name': 'elegy',
    'version': '0.7.4',
    'description': 'Elegy is a Neural Networks framework based on Jax and Haiku.',
    'long_description': '# Elegy\n[![PyPI Status Badge](https://badge.fury.io/py/elegy.svg)](https://pypi.org/project/elegy/)\n[![Coverage](https://img.shields.io/codecov/c/github/poets-ai/elegy?color=%2334D058)](https://codecov.io/gh/poets-ai/elegy)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/elegy)](https://pypi.org/project/elegy/)\n[![Documentation](https://img.shields.io/badge/api-reference-blue.svg)](https://poets-ai.github.io/elegy/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Contributions welcome](https://img.shields.io/badge/contributions-welcome-brightgreen.svg?style=flat)](https://github.com/poets-ai/elegy/issues)\n[![Status](https://github.com/poets-ai/elegy/workflows/GitHub%20CI/badge.svg)](https://github.com/poets-ai/elegy/actions?query=workflow%3A"GitHub+CI")\n\n-----------------\n\n_Elegy is a framework-agnostic Trainer interface for the Jax ecosystem._  \n\n#### Main Features\n* **Easy-to-use**: Elegy provides a Keras-like high-level API that makes it very easy to do common tasks.\n* **Flexible**: Elegy provides a functional Pytorch Lightning-like low-level API that provides maximal flexibility when needed.\n* **Agnostic**: Elegy supports a variety of frameworks including Flax, Haiku, and Optax on the high-level API, and it is 100% framework-agnostic on the low-level API.\n* **Compatible**: Elegy can consume a wide variety of common data sources including TensorFlow Datasets, Pytorch DataLoaders, Python generators, and Numpy pytrees.\n\nFor more information take a look at the [Documentation](https://poets-ai.github.io/elegy).\n\n## Installation\n\nInstall Elegy using pip:\n```bash\npip install elegy\n```\n\nFor Windows users we recommend the Windows subsystem for linux 2 [WSL2](https://docs.microsoft.com/es-es/windows/wsl/install-win10?redirectedfrom=MSDN) since [jax](https://github.com/google/jax/issues/438) does not support it yet.\n\n## Quick Start: High-level API\nElegy\'s high-level API provides a very simple interface you can use by implementing following steps:\n\n**1.** Define the architecture inside a `Module`. We will use Flax Linen for this example:\n```python\nimport flax.linen as nn\nimport jax\n\nclass MLP(nn.Module):\n    @nn.compact\n    def __call__(self, x):\n        x = nn.Dense(300)(x)\n        x = jax.nn.relu(x)\n        x = nn.Dense(10)(x)\n        return x\n```\n\n**2.** Create a `Model` from this module and specify additional things like losses, metrics, and optimizers:\n```python\nimport elegy, optax\n\nmodel = elegy.Model(\n    module=MLP(),\n    loss=[\n        elegy.losses.SparseCategoricalCrossentropy(from_logits=True),\n        elegy.regularizers.GlobalL2(l=1e-5),\n    ],\n    metrics=elegy.metrics.SparseCategoricalAccuracy(),\n    optimizer=optax.rmsprop(1e-3),\n)\n```\n**3.** Train the model using the `fit` method:\n```python\nmodel.fit(\n    x=X_train,\n    y=y_train,\n    epochs=100,\n    steps_per_epoch=200,\n    batch_size=64,\n    validation_data=(X_test, y_test),\n    shuffle=True,\n    callbacks=[elegy.callbacks.TensorBoard("summaries")]\n)\n```\n\n## Quick Start: Low-level API\nIn Elegy\'s low-level API lets you define exactly what goes on during training, testing, and inference. Lets define the `test_step` to implement a linear classifier in pure jax:\n\n**1.** Calculate our loss, logs, and states:\n```python\nclass LinearClassifier(elegy.Model):\n    # request parameters by name via depending injection.\n    # names: x, y_true, sample_weight, class_weight, states, initializing\n    def test_step(\n        self,\n        x, # inputs\n        y_true, # labels\n        states: elegy.States, # model state\n        initializing: bool, # if True we should initialize our parameters\n    ):  \n        rng: elegy.RNGSeq = states.rng\n        # flatten + scale\n        x = jnp.reshape(x, (x.shape[0], -1)) / 255\n        # initialize or use existing parameters\n        if initializing:\n            w = jax.random.uniform(\n                rng.next(), shape=[np.prod(x.shape[1:]), 10]\n            )\n            b = jax.random.uniform(rng.next(), shape=[1])\n        else:\n            w, b = states.net_params\n        # model\n        logits = jnp.dot(x, w) + b\n        # categorical crossentropy loss\n        labels = jax.nn.one_hot(y_true, 10)\n        loss = jnp.mean(-jnp.sum(labels * jax.nn.log_softmax(logits), axis=-1))\n        accuracy=jnp.mean(jnp.argmax(logits, axis=-1) == y_true)\n        # metrics\n        logs = dict(\n            accuracy=accuracy,\n            loss=loss,\n        )\n        return loss, logs, states.update(net_params=(w, b))\n```\n\n**2.** Instantiate our `LinearClassifier` with an optimizer:\n```python\nmodel = LinearClassifier(\n    optimizer=optax.rmsprop(1e-3),\n)\n```\n**3.** Train the model using the `fit` method:\n```python\nmodel.fit(\n    x=X_train,\n    y=y_train,\n    epochs=100,\n    steps_per_epoch=200,\n    batch_size=64,\n    validation_data=(X_test, y_test),\n    shuffle=True,\n    callbacks=[elegy.callbacks.TensorBoard("summaries")]\n)\n```\n#### Using Jax Frameworks\nIt is straightforward to integrate other functional JAX libraries with this \nlow-level API:\n\n```python\nclass LinearClassifier(elegy.Model):\n    def test_step(\n        self, x, y_true, states: elegy.States, initializing: bool\n    ):\n        rng: elegy.RNGSeq = states.rng\n        x = jnp.reshape(x, (x.shape[0], -1)) / 255\n        if initializing:\n            logits, variables = self.module.init_with_output(\n                {"params": rng.next(), "dropout": rng.next()}, x\n            )\n        else:\n            variables = dict(params=states.net_params, **states.net_states)\n            logits, variables = self.module.apply(\n                variables, x, rngs={"dropout": rng.next()}, mutable=True\n            )\n        net_states, net_params = variables.pop("params")\n        \n        labels = jax.nn.one_hot(y_true, 10)\n        loss = jnp.mean(-jnp.sum(labels * jax.nn.log_softmax(logits), axis=-1))\n        accuracy = jnp.mean(jnp.argmax(logits, axis=-1) == y_true)\n\n        logs = dict(accuracy=accuracy, loss=loss)\n        return loss, logs, states.update(net_params=net_params, net_states=net_states)\n```\n\n## More Info\n* [Getting Started: High-level API](https://poets-ai.github.io/elegy/getting-started-high-level-api/) tutorial.\n* [Getting Started: Low-level API](https://poets-ai.github.io/elegy/getting-started-low-level-api/) tutorial.\n* Elegy\'s [Documentation](https://poets-ai.github.io/elegy).\n* The [examples](https://github.com/poets-ai/elegy/tree/master/examples) directory.\n* [What is Jax?](https://github.com/google/jax#what-is-jax)\n\n### Examples\nTo run the examples first install some required packages:\n```\npip install -r examples/requirements.txt\n```\nNow run the example:\n```\npython examples/flax_mnist_vae.py \n```\n\n## Contributing\nDeep Learning is evolving at an incredible pace, there is so much to do and so few hands. If you wish to contribute anything from a loss or metric to a new awesome feature for Elegy just open an issue or send a PR! For more information check out our [Contributing Guide](https://poets-ai.github.io/elegy/guides/contributing).\n\n## About Us\nWe are some friends passionate about ML.\n\n## License\nApache\n\n## Citing Elegy\n\nTo cite this project:\n\n**BibTeX**\n\n```\n@software{elegy2020repository,\nauthor = {PoetsAI},\ntitle = {Elegy: A framework-agnostic Trainer interface for the Jax ecosystem},\nurl = {https://github.com/poets-ai/elegy},\nversion = {0.7.4},\nyear = {2020},\n}\n```\n\n\nWhere the current *version* may be retrieved either from the `Release` tag or the file [elegy/\\_\\_init\\_\\_.py](https://github.com/poets-ai/elegy/blob/master/elegy/__init__.py) and the *year* corresponds to the project\'s release year.\n',
    'author': 'Cristian Garcia',
    'author_email': 'cgarcia.e88@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://poets-ai.github.io/elegy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<3.9',
}


setup(**setup_kwargs)



import typing as tp
from abc import ABC, abstractmethod

from elegy import types, utils

REGISTRY: tp.Dict[tp.Type, tp.Type["GeneralizedModule"]] = {}


class ModuleExists(Exception):
    pass


class GeneralizedModule(ABC):
    @abstractmethod
    def __init__(self, module: tp.Any):
        ...

    @abstractmethod
    def init(self, rng: types.RNGSeq) -> tp.Callable[..., types.OutputStates]:
        ...

    @abstractmethod
    def apply(
        self,
        params: tp.Any,
        states: tp.Any,
        training: bool,
        rng: types.RNGSeq,
    ) -> tp.Callable[..., types.OutputStates]:
        ...

    def get_summary_params(
        self,
        path: types.Path,
        module: tp.Any,
        value: tp.Any,
        net_params: types.NetParams,
        net_states: types.NetStates,
    ) -> tp.Tuple[tp.Optional[types.Pytree], tp.Optional[types.Pytree]]:
        return None, None

    def update(
        self,
        params: tp.Optional[types.ModuleParams],
        states: tp.Optional[types.ModuleStates],
    ):
        pass


class CallableModule(GeneralizedModule):
    def __init__(self, f: tp.Callable):
        self.f = f

    def init(self, rng: types.RNGSeq) -> tp.Callable[..., types.OutputStates]:
        def lambda_(*args, **kwargs) -> types.OutputStates:

            output = utils.inject_dependencies(self.f)(*args, **kwargs)

            if isinstance(output, types.OutputStates):
                return output
            else:
                return types.OutputStates(
                    preds=output,
                    params=None,
                    states=None,
                )

        return lambda_

    def apply(
        self,
        params: tp.Any,
        states: tp.Any,
        training: bool,
        rng: types.RNGSeq,
    ) -> tp.Callable[..., types.OutputStates]:
        def lambda_(*args, **kwargs) -> types.OutputStates:

            output = utils.inject_dependencies(self.f)(*args, **kwargs)

            if isinstance(output, types.OutputStates):
                return output
            else:
                return types.OutputStates(
                    preds=output,
                    params=None,
                    states=None,
                )

        return lambda_


def register_module_for(
    module_type,
) -> tp.Callable[[tp.Type[GeneralizedModule]], tp.Any]:
    def wrapper(generalized_module_type: tp.Type[GeneralizedModule]) -> tp.Any:
        if module_type in REGISTRY:
            raise ModuleExists(
                f"Type {module_type} already registered with {REGISTRY[module_type]}"
            )

        REGISTRY[module_type] = generalized_module_type

        return generalized_module_type

    return wrapper


def generalize(
    module: tp.Any,
    callable_default: tp.Type[GeneralizedModule] = CallableModule,
) -> GeneralizedModule:

    if isinstance(module, GeneralizedModule):
        return module

    generalized_module_type: tp.Optional[tp.Type[GeneralizedModule]] = None

    for module_type in REGISTRY:
        if isinstance(module, module_type):
            generalized_module_type = REGISTRY[module_type]

    if generalized_module_type is None:
        if isinstance(module, tp.Callable):
            return callable_default(module)
        else:
            raise ValueError(f"No GeneralizedModule found for {module}.")

    return generalized_module_type(module)


def is_generalizable(module: tp.Any, accept_callable: bool = False) -> bool:

    if isinstance(module, GeneralizedModule):
        return True

    for module_type in REGISTRY:
        if isinstance(module, module_type):
            return True

    if accept_callable and isinstance(module, tp.Callable):
        return True

    return False



from elegy import types
import typing as tp

import jax.numpy as jnp

from elegy.metrics.reduce import Reduce, Reduction


class Sum(Reduce):
    """
    Computes the (weighted) sum of the given values.

    For example, if values is [1, 3, 5, 7] then the sum is 16.
    If the weights were specified as [1, 1, 0, 0] then the sum would be 4.
    This metric creates one variable, `total`, that is used to compute the sum of
    `values`. This is ultimately returned as `sum`.
    If `sample_weight` is `None`, weights default to 1.  Use `sample_weight` of 0
    to mask values.

    Usage:
    ```python
    m = elegy.metrics.Sum()
    assert 16.0 == m([1, 3, 5, 7])
    ```
    Usage with Elegy API:
    ```python
    model = elegy.Model(
        module_fn,
        loss=elegy.losses.CategoricalCrossentropy(),
        metrics=elegy.metrics.Sum(name='sum_1'),
    )
    model = elegy.Model(inputs, outputs)
    ```
    """

    def __init__(self, on: tp.Optional[types.IndexLike] = None, **kwargs):
        """Creates a `Sum` instance.

        Arguments:
            on: A string or integer, or iterable of string or integers, that
                indicate how to index/filter the `y_true` and `y_pred`
                arguments before passing them to `call`. For example if `on = "a"` then
                `y_true = y_true["a"]`. If `on` is an iterable
                the structures will be indexed iteratively, for example if `on = ["a", 0, "b"]`
                then `y_true = y_true["a"][0]["b"]`, same for `y_pred`. For more information
                check out [Keras-like behavior](https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#keras-like-behavior).
            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(reduction=Reduction.SUM, on=on, **kwargs)



from .resnet import (
    ResNet,
    ResNet18,
    ResNet34,
    ResNet50,
    ResNet101,
    ResNet152,
    ResNet200,
)

__all__ = [
    "ResNet",
    "ResNet18",
    "ResNet34",
    "ResNet50",
    "ResNet101",
    "ResNet152",
    "ResNet200",
]



import elegy


import jax, jax.numpy as jnp
import numpy as np
import tensorflow.keras as keras


#
def test_basic():

    y_true = jnp.array([1, 2])
    y_pred = jnp.array([[0.05, 0.95, 0], [0.1, 0.8, 0.1]])

    # Using 'auto'/'sum_over_batch_size' reduction type.
    scce = elegy.losses.SparseCategoricalCrossentropy()
    result = scce(y_true, y_pred)  # 1.177
    assert jnp.isclose(result, 1.177, rtol=0.01)

    # Calling with 'sample_weight'.
    result = scce(y_true, y_pred, sample_weight=jnp.array([0.3, 0.7]))  # 0.814
    assert jnp.isclose(result, 0.814, rtol=0.01)

    # Using 'sum' reduction type.
    scce = elegy.losses.SparseCategoricalCrossentropy(
        reduction=elegy.losses.Reduction.SUM
    )
    result = scce(y_true, y_pred)  # 2.354
    assert jnp.isclose(result, 2.354, rtol=0.01)

    # Using 'none' reduction type.
    scce = elegy.losses.SparseCategoricalCrossentropy(
        reduction=elegy.losses.Reduction.NONE
    )
    result = scce(y_true, y_pred)  # [0.0513, 2.303]
    assert jnp.all(jnp.isclose(result, [0.0513, 2.303], rtol=0.01))


def test_scce_out_of_bounds():
    ypred = jnp.zeros([4, 10])
    ytrue0 = jnp.array([0, 0, -1, 0])
    ytrue1 = jnp.array([0, 0, 10, 0])

    scce = elegy.losses.SparseCategoricalCrossentropy()

    assert jnp.isnan(scce(ytrue0, ypred)).any()
    assert jnp.isnan(scce(ytrue1, ypred)).any()

    scce = elegy.losses.SparseCategoricalCrossentropy(check_bounds=False)
    assert not jnp.isnan(scce(ytrue0, ypred)).any()
    assert not jnp.isnan(scce(ytrue1, ypred)).any()


def test_scce_uint8_ytrue():
    ypred = np.random.random([2, 256, 256, 10])
    ytrue = np.random.randint(0, 10, size=(2, 256, 256)).astype(np.uint8)

    loss0 = elegy.losses.sparse_categorical_crossentropy(ytrue, ypred, from_logits=True)
    loss1 = keras.losses.sparse_categorical_crossentropy(ytrue, ypred, from_logits=True)

    assert np.allclose(loss0, loss1)



import functools
import typing as tp

import haiku as hk
import numpy as np

from elegy import utils
from elegy import hooks, module


def sequential(*layers: tp.Callable[..., tp.Any]) -> tp.Callable[..., tp.Any]:
    """
    Connects all layers. `*args` and `**kwargs` are passed to the first layer.

    ```python
    def call(self, x):
        mlp = elegy.nn.sequential(
            elegy.nn.Linear(64),
            jax.nn.relu,
            elegy.nn.Linear(32),
            jax.nn.relu,
            elegy.nn.Linear(10),
            jax.nn.softmax,
        )
        y = mlp(x)
        ...
    ```

    !!! Note
        `sequential` is not a `Module`, that is, it wont create a scope over the layers it runs,
        in contrast to `Sequential` layers are eagerly instantiate outside of `sequential`
        and just passed to it to automate the execution.

    Arguments:
        layers: Modules or functions passed as `*args`

    Returns:
        A callable that waits for the inputs and applies the layers sequentially.
    """

    def call(inputs, *args, **kwargs):

        out = inputs
        for i, layer in enumerate(layers):
            if i == 0:
                out = layer(out, *args, **kwargs)
            else:
                out = layer(out)

            if not isinstance(layer, module.Module):
                if hooks.summaries_active():
                    name = utils.get_name(layer)

                    path = module.get_module_path()
                    path = path if path is not None else ()

                    hooks.add_summary(path + (name,), layer, out)
        return out

    return call


class Sequential(module.Module):
    """
    Creates a Module from a zero argument lambda that produces a list of Modules or function to be executed sequentially. The lambda is necessary so that all sub-modules are instantiated inside the context of the Sequential module.

    ```python
    mlp = elegy.nn.Sequential(
        lambda: [
            elegy.nn.Linear(64),
            jax.nn.relu,
            elegy.nn.Linear(32),
            jax.nn.relu,
            elegy.nn.Linear(10),
            jax.nn.softmax,
        ]
    )
    y = mlp(x)
    ```
    """

    def __init__(
        self, layers: tp.Callable[[], tp.Iterable[tp.Callable[..., tp.Any]]], **kwargs
    ):
        super().__init__(**kwargs)
        self.layers = tuple(layers())

        if len(self.layers) > 0:
            self._signature_f = (
                self.layers[0]._signature_f
                if hasattr(self.layers[0], "_signature_f")
                else self.layers[0]
            )

    def call(self, *args, **kwargs):
        """Connects all layers. `*args` and `**kwargs` are passed to the first layer."""
        return sequential(*self.layers)(*args, **kwargs)



from elegy import types
from elegy import utils
import typing as tp

import jax.numpy as jnp

from elegy.metrics.reduce_confusion_matrix import Reduction, ReduceConfusionMatrix
from elegy.metrics.metric import Metric


def recall(
    y_true: jnp.ndarray,
    y_pred: jnp.ndarray,
    threshold: jnp.ndarray,
    class_id: jnp.ndarray,
    sample_weight: jnp.ndarray,
    true_positives: ReduceConfusionMatrix,
    false_negatives: ReduceConfusionMatrix,
) -> jnp.ndarray:

    # TODO: class_id behavior
    y_pred = (y_pred > threshold).astype(jnp.float32)

    if y_true.dtype != y_pred.dtype:
        y_pred = y_pred.astype(y_true.dtype)

    true_positives = true_positives(
        y_true=y_true, y_pred=y_pred, sample_weight=sample_weight
    )
    false_negatives = false_negatives(
        y_true=y_true, y_pred=y_pred, sample_weight=sample_weight
    )

    return jnp.nan_to_num(jnp.divide(true_positives, true_positives + false_negatives))


class Recall(Metric):
    """
    This metric creates two local variables, `true_positives` and
    `false_negatives`, that are used to compute the recall. This value is
    ultimately returned as `recall`, an idempotent operation that simply divides
    `true_positives` by the sum of `true_positives` and `false_negatives`.

    If `sample_weight` is `None`, weights default to 1. Use `sample_weight` of 0 to mask values.

    If sample_weight is None, weights default to 1. Use sample_weight of 0 to mask values.

    If class_id is specified, we calculate recall by considering only the entries in the batch
    for which class_id is above the threshold and computing the fraction of them for which class_id
    is indeed a correct label.

    ```python
        recall = elegy.metrics.Recall()

        result = recall(
            y_true=jnp.array([0, 1, 1, 1]), y_pred=jnp.array([1, 0, 1, 1])
        )
        assert result == 0.6666667 # 2 / 3

        result = recall(
            y_true=jnp.array([1, 1, 1, 1]), y_pred=jnp.array([1, 0, 0, 0])
        )
        assert result == 0.42857143 # 3 / 7
    ```

    Usage with elegy API:

    ```python
        model = elegy.Model(
        module_fn,
        loss=elegy.losses.CategoricalCrossentropy(),
        metrics=elegy.metrics.Recall(),
        optimizer=optix.adam(1e-3),
    )
    ```
    """

    def __init__(
        self,
        on: tp.Optional[types.IndexLike] = None,
        threshold=None,
        class_id=None,
        **kwargs
    ):
        """
        Creates a `Recall` instance.

        Arguments:
            on: A string or integer, or iterable of string or integers, that
                indicate how to index/filter the `y_true` and `y_pred`
                arguments before passing them to `call`. For example if `on = "a"` then
                `y_true = y_true["a"]`. If `on` is an iterable
                the structures will be indexed iteratively, for example if `on = ["a", 0, "b"]`
                then `y_true = y_true["a"][0]["b"]`, same for `y_pred`. For more information
                check out [Keras-like behavior](https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#keras-like-behavior).

            threshold: (Optional) A float value or a python list/tuple of float threshold
                values in [0, 1]. A threshold is compared with prediction values to determine
                the truth value of predictions (i.e., above the threshold is true, below is false).
                One metric value is generated for each threshold value. If neither threshold is set
                the default is to calculate recall with threshold=0.5.

            class_id: (Optional) Integer class ID for which we want binary metrics.
                This must be in the half-open interval `[0, num_classes)`, where
                `num_classes` is the last dimension of predictions.

            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(on=on, **kwargs)
        self.threshold = 0.5 if threshold is None else threshold
        self.class_id = 1 if class_id is None else class_id
        self.true_positives = ReduceConfusionMatrix(reduction=Reduction.TRUE_POSITIVES)
        self.false_negatives = ReduceConfusionMatrix(
            reduction=Reduction.FALSE_NEGATIVES
        )

    def call(
        self,
        y_true: jnp.ndarray,
        y_pred: jnp.ndarray,
        sample_weight: tp.Optional[jnp.ndarray] = None,
    ) -> jnp.ndarray:
        """
        Accumulates metric statistics. `y_true` and `y_pred` should have the same shape.

        Arguments:
            y_true: Ground truth values. shape = `[batch_size, d0, .. dN]`.

            y_pred: The predicted values. shape = `[batch_size, d0, .. dN]`.

            sample_weight: Optional weighting of each example. Defaults to 1. Can be a
                `Tensor` whose rank is either 0, or the same rank as `y_true`, and must
                be broadcastable to `y_true`.
        Returns:
            Array with the cumulative recall.
        """

        return recall(
            y_true=y_true,
            y_pred=y_pred,
            sample_weight=sample_weight,
            threshold=self.threshold,
            class_id=self.class_id,
            true_positives=self.true_positives,
            false_negatives=self.false_negatives,
        )



import typing as tp
from enum import Enum

import jax.numpy as jnp
import numpy as np

from elegy import initializers, module, types, utils, module
from elegy.metrics.metric import Metric


class Reduction(Enum):
    TRUE_POSITIVES = "true_positives"
    FALSE_POSITIVES = "false_positives"
    FALSE_NEGATIVES = "false_negatives"
    TRUE_NEGATIVES = "true_negatives"


def reduce(
    cm_metric: jnp.ndarray,
    y_true: jnp.ndarray,
    y_pred: jnp.ndarray,
    reduction: Reduction,
    sample_weight: tp.Optional[np.ndarray],
    dtype: jnp.dtype,
) -> tp.Tuple[jnp.ndarray, jnp.ndarray, tp.Optional[jnp.ndarray]]:

    if sample_weight is not None:
        sample_weight = sample_weight.astype(dtype)

        # Update dimensions of weights to match with values if possible.
        # values, _, sample_weight = tf_losses_utils.squeeze_or_expand_dimensions(
        #     values, sample_weight=sample_weight
        # )

        try:
            # Broadcast weights if possible.
            sample_weight = jnp.broadcast_to(sample_weight, y_true.shape)
        except ValueError as e:
            raise e

    if reduction == Reduction.TRUE_POSITIVES:
        mask = y_pred == 1
        hits = y_true == y_pred
        if sample_weight is not None:
            hits = hits * sample_weight
        value = jnp.sum(hits * mask)

    if reduction == Reduction.FALSE_POSITIVES:
        mask = y_pred == 1
        misses = y_true != y_pred
        if sample_weight is not None:
            misses = misses * sample_weight
        value = jnp.sum(misses * mask)

    if reduction == Reduction.FALSE_NEGATIVES:
        mask = y_true == 1
        misses = y_true != y_pred
        if sample_weight is not None:
            misses = misses * sample_weight
        value = jnp.sum(misses * mask)

    if reduction == Reduction.TRUE_NEGATIVES:
        mask = y_true == 0
        hits = y_true == y_pred
        if sample_weight is not None:
            hits = hits * sample_weight
        value = jnp.sum(hits * mask)

    cm_metric += value
    return cm_metric


class ReduceConfusionMatrix(Metric):
    """Encapsulates confusion matrix metrics that perform a reduce operation on the values."""

    def __init__(
        self, reduction: Reduction, on: tp.Optional[types.IndexLike] = None, **kwargs
    ):
        super().__init__(on=on, **kwargs)

        self._reduction = reduction

        if self._reduction not in (
            Reduction.TRUE_POSITIVES,
            Reduction.FALSE_POSITIVES,
            Reduction.FALSE_NEGATIVES,
            Reduction.TRUE_NEGATIVES,
        ):
            raise NotImplementedError(
                "reduction {reduction} not implemented".format(
                    reduction=self._reduction
                )
            )

    def call(
        self,
        y_true: jnp.ndarray,
        y_pred: jnp.ndarray,
        sample_weight: tp.Optional[jnp.ndarray] = None,
    ) -> tp.Any:
        """
        Accumulates confusion matrix metrics for computing the reduction metric.

        Arguments:
            y_true: Ground truth values. shape = `[batch_size, d0, .. dN]`.

            y_pred: The predicted values. shape = `[batch_size, d0, .. dN]`.

            sample_weight: Optional weighting of each example. Defaults to 1.

        Returns:
            Array with the cumulative reduce metric.
        """

        cm_metric = self.add_parameter(
            "cm_metric",
            lambda: jnp.array(0, dtype=jnp.int32),
            trainable=False,
        )

        cm_metric = reduce(
            cm_metric=cm_metric,
            y_true=y_true,
            y_pred=y_pred,
            reduction=self._reduction,
            sample_weight=sample_weight,
            dtype=self.dtype,
        )

        self.update_parameter("cm_metric", cm_metric)

        return cm_metric



import elegy


import jax.numpy as jnp
import jax
import tensorflow.keras as tfk

# import debugpy

# print("Waiting for debugger...")
# debugpy.listen(5679)
# debugpy.wait_for_client()


def test_basic():

    y_true = jnp.array([[0, 1], [0, 0]])
    y_pred = jnp.array([[0.6, 0.4], [0.4, 0.6]])

    # Using 'auto'/'sum_over_batch_size' reduction type.
    huber_loss = elegy.losses.Huber()
    assert huber_loss(y_true, y_pred) == 0.155

    # Calling with 'sample_weight'.
    assert huber_loss(y_true, y_pred, sample_weight=jnp.array([0.8, 0.2])) == 0.08500001

    # Using 'sum' reduction type.
    huber_loss = elegy.losses.Huber(reduction=elegy.losses.Reduction.SUM)
    assert huber_loss(y_true, y_pred) == 0.31

    # Using 'none' reduction type.
    huber_loss = elegy.losses.Huber(reduction=elegy.losses.Reduction.NONE)

    assert jnp.equal(huber_loss(y_true, y_pred), jnp.array([0.18, 0.13000001])).all()


def test_function():

    rng = jax.random.PRNGKey(42)

    y_true = jax.random.randint(rng, shape=(2, 3), minval=0, maxval=2)
    y_pred = jax.random.uniform(rng, shape=(2, 3))

    loss = elegy.losses.huber(y_true, y_pred, delta=1.0)
    assert loss.shape == (2,)

    y_pred = y_pred.astype(float)
    y_true = y_true.astype(float)
    delta = 1.0
    error = jnp.subtract(y_pred, y_true)
    abs_error = jnp.abs(error)
    quadratic = jnp.minimum(abs_error, delta)
    linear = jnp.subtract(abs_error, quadratic)
    assert jnp.array_equal(
        loss,
        jnp.mean(
            jnp.add(
                jnp.multiply(0.5, jnp.multiply(quadratic, quadratic)),
                jnp.multiply(delta, linear),
            ),
            axis=-1,
        ),
    )


def test_compatibility():
    # Input:  true (y_true) and predicted (y_pred) tensors
    rng = jax.random.PRNGKey(121)

    y_true = jax.random.randint(rng, shape=(2, 3), minval=0, maxval=2)
    y_true = y_true.astype(dtype=jnp.float32)
    y_pred = jax.random.uniform(rng, shape=(2, 3))

    # cosine_loss using sample_weight
    huber_loss = elegy.losses.Huber(delta=1.0)
    huber_loss_tfk = tfk.losses.Huber(delta=1.0)

    assert jnp.isclose(
        huber_loss(y_true, y_pred, sample_weight=jnp.array([1, 0])),
        huber_loss_tfk(y_true, y_pred, sample_weight=jnp.array([1, 0])),
        rtol=0.0001,
    )

    # cosine_loss with reduction method: SUM
    huber_loss = elegy.losses.Huber(delta=1.0, reduction=elegy.losses.Reduction.SUM)
    huber_loss_tfk = tfk.losses.Huber(delta=1.0, reduction=tfk.losses.Reduction.SUM)
    assert jnp.isclose(
        huber_loss(y_true, y_pred), huber_loss_tfk(y_true, y_pred), rtol=0.0001
    )

    # cosine_loss with reduction method: NONE
    huber_loss = elegy.losses.Huber(delta=1.0, reduction=elegy.losses.Reduction.NONE)
    huber_loss_tfk = tfk.losses.Huber(delta=1.0, reduction=tfk.losses.Reduction.NONE)
    assert jnp.all(
        jnp.isclose(
            huber_loss(y_true, y_pred), huber_loss_tfk(y_true, y_pred), rtol=0.0001
        )
    )


if __name__ == "__main__":

    test_basic()
    test_function()
    test_compatibility()
